import json
from time import time

import config
from uchan import g
from uchan.lib.utils import now
from werkzeug.contrib.cache import MemcachedCache


def make_attr_dict(value):
    if isinstance(value, list):
        value = [make_attr_dict(i) for i in value]
    elif isinstance(value, dict):
        if not isinstance(value, CacheDict):
            value = CacheDict(value)
        for key in value:
            value[key] = make_attr_dict(value[key])
    return value


class CacheWrapper(MemcachedCache):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = self._client

        self.max_length = config.MEMCACHE_MAX_ITEM_SIZE
        self.client.server_max_value_length = self.max_length

    def set(self, key, value, **kwargs):
        # g.logger.debug('set {} {}'.format(key, value))

        json_data = json.dumps(value, separators=(',', ':'))

        if len(json_data) > self.max_length:
            g.logger.error('cache value exceeds max length ({} > {})'.format(len(json_data), self.max_length))
            return False

        percentage = len(json_data) / self.max_length
        if percentage > 0.5:
            g.logger.warning('key {0} exceeds 50% of the total storage available ({1:.2f}%)'.format(key, percentage * 100))

        ret = super().set(key, json_data, **kwargs)
        if not ret:
            g.logger.error('cache set failed {}'.format(ret))
        return bool(ret)

    def get(self, key, convert=False):
        # g.logger.debug('get {}'.format(key))
        res = super().get(key)
        if res is None:
            return None
        else:
            data = json.loads(res)
            if convert:
                return make_attr_dict(data)
            else:
                return data

    def delete(self, key):
        # logger.debug('delete {}'.format(key))
        super().delete(key)

    def _normalize_timeout(self, timeout):
        if timeout is None:
            return self.default_timeout
        # Allow zero to mean the same as does not expire
        if timeout == 0:
            return 0
        return int(time()) + timeout


class CacheDict(dict):
    """
    Makes keys accessible like a property
    """

    def __init__(self, *args, **kwargs):
        try:
            super().__init__(*args, **kwargs)
        except ValueError:
            pass
        self.__dict__ = self

    def convert(self):
        return make_attr_dict(self)


class LocalCache:
    """
    Super simple local cache with no pruning based on a dict.
    Don't do anything special.
    """

    def __init__(self):
        self.items = {}

    def set(self, key, value, timeout=15000):
        self.items[key] = (now() + timeout, value)

    def get(self, key):
        try:
            expires, item = self.items[key]
            if now() < expires:
                return item
        except KeyError:
            pass
        return None


from uchan.lib.cache.board_cache import BoardCache
from uchan.lib.cache.posts_cache import PostsCache
from uchan.lib.cache.site_cache import SiteCache
from uchan.lib.cache.page_cache import PageCache
